#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-3.0-or-later

# gui - Graphical User Interface module.

# Copyright (C) 2022 Sergio Chica Manjarrez @ pervasive.it.uc3m.es.
# Universidad Carlos III de Madrid.

# This file is part of neuropy3.

# neuropy3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# neuropy3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with GNU Emacs.  If not, see <https://www.gnu.org/licenses/>.

from PySide6.QtCharts import (QBarSet, QCategoryAxis, QLineSeries,
                              QSplineSeries, QValueAxis)
from PySide6.QtCore import QObject, QPointF, Signal, Slot
from PySide6.QtQml import QQmlApplicationEngine
from PySide6.QtWidgets import QApplication
from neuropy3.neuropy3 import MindWave
from threading import Event, Thread
from neuropy3.gui import resources  # noqa
from PySide6.QtGui import QIcon
from queue import Queue

import neuropy3.utils as ut
import numpy as np
import math
import sys


class BackendThread(Thread):
    def __init__(self, root, backend, flag, address):
        Thread.__init__(self)
        self.root = root
        self.backend = backend
        self.flag = flag
        self.address = address
        self.queue = Queue()

    def run(self):
        self.mindwave = MindWave(address=self.address, autostart=False,
                                 verbose=2)
        ut.set_logger(self.root.newLineConsole)
        self.mindwave.update_callback('eeg', self.send_eeg)
        self.mindwave.update_callback('raw', self.send_raw)
        self.mindwave.update_callback('attention', self.send_attention)
        self.mindwave.update_callback('meditation', self.send_meditation)
        try:
            self.mindwave.start()
        except SystemExit:
            self.root.enableStartButton.emit()
            sys.exit()
        self.flag.wait()
        self.mindwave.stop()

    def send_eeg(self, eeg):
        if 0 in eeg.values():
            return
        try:
            data = {band: math.log(value) for band, value in eeg.items()}
            self.backend.update_asic(data)
        except ValueError:
            pass

    def send_raw(self, raw):
        self.queue.put(raw)
        if self.queue.qsize() > ut.SAMPLE_RATE:
            microvolts = [ut.raw_to_microvolt(self.queue.get())
                          for d in range(ut.SAMPLE_RATE)]
            self.backend.update_raw(microvolts)

    def send_attention(self, att):
        self.root.attUpdate.emit(att)

    def send_meditation(self, med):
        self.root.medUpdate.emit(med)


class Backend(QObject):
    newChart = Signal(str, QLineSeries, QValueAxis)
    newPolar = Signal(QSplineSeries, QCategoryAxis)
    newBars = Signal(QBarSet, QBarSet,
                     QBarSet, QBarSet,
                     QBarSet, QBarSet,
                     QBarSet, QBarSet)

    def __init__(self):
        QObject.__init__(self)
        self.charts = {band: {'serie': None, 'axis': None}
                       for band in ut.EEG}
        self.charts['raw'] = {'serie': None, 'axis': None}
        self.asic = {band: None for band in ut.NAMES[6:]}
        self.polar = {band: idx for idx, band in enumerate(ut.NAMES[6:])}
        self.polar['serie'] = None
        self.time = np.arange(0, 1, 1/ut.SAMPLE_RATE)
        self.idx = 0

    @Slot(str, QLineSeries, QValueAxis)
    def store_new_chart(self, chart, serie, axis):
        self.charts[chart]['serie'] = serie
        self.charts[chart]['axis'] = axis

    @Slot(QSplineSeries, QCategoryAxis)
    def store_new_polar(self, serie, category):
        self.polar['serie'] = serie
        for idx, band in enumerate(self.asic):
            category.append(band.upper(), idx)

    @Slot(QBarSet, QBarSet, QBarSet, QBarSet,
          QBarSet, QBarSet, QBarSet, QBarSet)
    def store_new_bars(self, *bars):
        for band, bar in zip(ut.NAMES[6:], bars):
            self.asic[band] = bar

    def update_raw(self, microvolts):
        bands = ut.microvolts_to_bands(microvolts)
        points = {
            'raw': [QPointF(x, y) for x, y in zip(self.time, microvolts)],
            'delta': [QPointF(x, y) for x, y in zip(self.time, bands[0])],
            'theta': [QPointF(x, y) for x, y in zip(self.time, bands[1])],
            'alpha': [QPointF(x, y) for x, y in zip(self.time, bands[2])],
            'beta': [QPointF(x, y) for x, y in zip(self.time, bands[3])],
            'gamma': [QPointF(x, y) for x, y in zip(self.time, bands[4])]
        }
        minmax = {
            'raw': ut.signal_axes(microvolts),
            'delta': ut.signal_axes(bands[0]),
            'theta': ut.signal_axes(bands[1]),
            'alpha': ut.signal_axes(bands[2]),
            'beta': ut.signal_axes(bands[3]),
            'gamma': ut.signal_axes(bands[4])
        }
        for chart in points:
            self.charts[chart]['axis'].setMin(minmax[chart][0])
            self.charts[chart]['axis'].setMax(minmax[chart][1])
            self.charts[chart]['serie'].replace(points[chart])

    def update_asic(self, data):
        points = [QPointF(self.polar[band], data[band]) for band in data]
        points.append(QPointF(8, data['delta']))
        if self.idx == 0:
            self.polar['serie'].append(points)
        else:
            self.polar['serie'].replace(points)
        self.polar['serie'].setColor(
            self.asic[max(data, key=data.get)].color())
        self.idx += 1
        for band in data:
            self.asic[band].replace(0, data[band])


def main(address=None):
    thread_flag = None
    thread = None

    def thread_start():
        nonlocal thread_flag, thread
        thread_flag = Event()
        thread = BackendThread(main, backend, thread_flag, address)
        thread.start()

    def thread_quit():
        if thread_flag is not None and thread is not None:
            thread_flag.set()
            thread.join()

    app = QApplication(sys.argv)
    app.setWindowIcon(QIcon(":/icon"))
    engine = QQmlApplicationEngine()
    engine.quit.connect(app.quit)
    backend = Backend()
    backend.newChart.connect(backend.store_new_chart)
    backend.newPolar.connect(backend.store_new_polar)
    backend.newBars.connect(backend.store_new_bars)
    engine.rootContext().setContextProperty('backend', backend)
    engine.load('neuropy3/gui/gui.qml')
    main = engine.rootObjects()[0]
    main.startThread.connect(thread_start)
    main.closing.connect(thread_quit)

    sys.exit(app.exec())



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-3.0-or-later

# neuropy3 - Backend module.

# Copyright (C) 2022 Sergio Chica Manjarrez @ pervasive.it.uc3m.es.
# Universidad Carlos III de Madrid.

# This file is part of neuropy3.

# neuropy3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# neuropy3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with GNU Emacs.  If not, see <https://www.gnu.org/licenses/>.

from threading import Thread, Event
from neuropy3 import utils as ut

import bluetooth
import struct
import sys


class MindWaveReader(Thread):
    """Thread class running in background. It reads every packet
    sent by NeuroSky MindWave Mobile 2 and updated MindWave class
    data.
    :param data: Shared data with MindWave class, updated on packet received
    :type data: dict
    :param callbacks: Shared callbacks with MindWave class,
                      called on packet type received (if present)
    :type callbacks: dict
    :param flag: Event flag to stop thread on demand
    :type flag: threading.Event
    :param socket: Bluetooth socket
    :type socket: bluetooth.BluetoothSocket
    :param verbose: Verbose level
    :type verbose: int. Allowed values: 0-4
    """
    def __init__(self, data, callbacks, flag, socket, verbose):
        Thread.__init__(self)
        self.data = data
        self.callbacks = callbacks
        self.flag = flag
        self.socket = socket
        self.verbose = verbose
        self.step = 0
        self.new = []

    def run(self):
        """Starts the read thread loop"""
        while not self.flag.is_set():
            self._read_packet()

    def _read(self, bytes=1):
        """Reads bytes from bluetooth socket
        :param bytes: Number of bytes to read
        :type bytes: int, optional. Default: 1
        :return: A bytearray read from bluetooth socket
        :rtype: bytearray"""
        self.socket.settimeout(5)
        try:
            return self.socket.recv(bytes)
        except bluetooth.btcommon.BluetoothError:
            ut.log('error', "Bluetooth timed out. Check headset is on.",
                   self.verbose)
            sys.exit(1)

    def _b2i(self, value, bytes=1):
        """Converts bytes to integer
        Depending of the size of the bytes, integer can be signed
        or unsigned. More info in thinkgear communications protocol
        :param value: Value to be converted
        :type value: bytearray
        :param bytes: Number of bytes to be converted
        :type bytes: int, optional. Allowed: 1, 2, 4. Default: 1
        :return: Value after converting bytearray
        :rtype: int"""
        vtype = '>B'
        if bytes == 2:
            vtype = '>h'
        elif bytes == 4:
            vtype = '>I'
        return struct.unpack(vtype, value)[0]

    def _read_eeg(self, values):
        """Parses ASIC_EEG_POWER payload to eeg bands
        :param values: Packets to be parsed
        :type values: bytearray"""
        for idx, band in enumerate(ut.NAMES[6:]):
            self.data['values']['eeg'][band] = self._b2i(
                b'\x00' + b''.join(values[idx*3:idx*3+3]), 4)
        if 'eeg' in self.callbacks:
            self.callbacks['eeg'](self.data['values']['eeg'])

    def _valid_payload(self):
        """Checks payload length and checksum"""
        valid = (False, None, 0)
        plength = self._b2i(self._read())
        if plength >= ut.PLENGTH_MAX:
            ut.log('warn', "Packet length too large. Packet discarded.",
                   self.verbose)
        else:
            payload = []
            chksum = 0
            for idx in range(plength):
                packet = self._read()
                payload.append(packet)
                chksum += self._b2i(packet)
            checksum = self._b2i(self._read())
            chksum = ~chksum & 0xFF
            if chksum != checksum:
                ut.log('warn', "Checksum failed. Packet discarded.",
                       self.verbose)
            else:
                valid = (True, payload, plength)
        return valid

    def _read_packet(self):
        """Reads packets and stores in shared data dict"""
        # Synchronize on SYNC bytes
        if self._read() != ut.BYTE['sync']:
            return
        if self._read() != ut.BYTE['sync']:
            return
        valid, payload, plength = self._valid_payload()
        if valid:
            self.data['packets'] += 1
            idx = 0
            while idx < plength:
                known = True
                code = payload[idx]
                ut.log('succ', f"Reading packet: {code}", self.verbose)
                if code in (ut.BYTE['step1'], ut.BYTE['step2']):
                    self.step += 1
                    if self.step == 2:
                        ut.log('info',
                               "MindWave connection established.",
                               self.verbose)
                    return
                elif code < ut.BYTE['_max']:
                    idx += 1
                    value = payload[idx]
                    if ut.CODE[code] in ut.NAMES[:5]:
                        value = self._b2i(value)
                        self._update(ut.CODE[code], value)
                        if code == ut.BYTE['signal']:
                            if value == ut.NO_CONTACT:
                                ut.log('warn',
                                       "MindWave electrodes are not in "
                                       "contact with your skin.",
                                       self.verbose)
                            elif value > ut.NO_CONTACT:
                                ut.log('warn',
                                       "MindWave poor signal detected. "
                                       "Check electrodes or "
                                       "interferences.",
                                       self.verbose)
                    else:
                        known = False
                else:
                    idx += 1
                    vlength = self._b2i(payload[idx])
                    if code == ut.BYTE['raw']:
                        if vlength != ut.PKT_RAW_MAX:
                            ut.log('warn',
                                   f"RAW wrong number of bytes: "
                                   f"{vlength}. Expected: {ut.PKT_RAW_MAX}. "
                                   f"Packet discarded.",
                                   self.verbose)
                        else:
                            value = self._b2i(
                                b''.join(payload[idx+1:idx+1+vlength]), 2)
                            self._update('raw', value)
                    elif code == ut.BYTE['eeg']:
                        if vlength != ut.PKT_EEG_MAX:
                            ut.log('warn',
                                   f"EEG wrong number of bytes: "
                                   f"{vlength}. Expected: {ut.PKT_EEG_MAX}. "
                                   f"Packet discarded.",
                                   self.verbose)
                        else:
                            self._read_eeg(payload[idx+1:idx+1+vlength])
                    else:
                        known = False
                    idx += vlength
                if not known:
                    ut.log('warn',
                           f"Code not recognized: {code}. "
                           f"Packet discarded.",
                           self.verbose)
                idx += 1

    def _update(self, name, value):
        """Updates shared data value and executes callback if present
        :param name: Name of variable to be updated
        :type name: str
        :param value: New value
        :type value: int"""
        self.data['values'][name] = value
        if name in self.callbacks:
            self.callbacks[name](value)


class MindWave:
    """Main class. Interface to interact with read data from NeuroSky
    MindWave Mobile 2
    :param address: Bluetooth address of NeuroSky MindWave Mobile 2
    :type address: str, optional
    :param autostart: Starts automatically MindWaveReader
    :type autostart: bool, optional. Default: True
    :param verbose: Verbose level
    :type verbose: int, optional. Allowed values: 0-4. Default: 1
    """
    def __init__(self, address=None, autostart=True, verbose=1):
        self.address = address
        self.verbose = verbose
        self._data = {
            'packets': 0,
            'values': {
                'signal': 0, 'attention': 0, 'meditation': 0, 'raw': 0,
                'eeg': {
                    'delta': 0, 'theta': 0,
                    'alpha_l': 0, 'alpha_h': 0,
                    'beta_l': 0, 'beta_h': 0,
                    'gamma_l': 0, 'gamma_m': 0
                }
            }
        }
        self.callbacks = {}
        self.thread = None
        self.socket = None
        self.flag = Event()
        if autostart:
            self.start()

    def scan(self):
        """Scans bluetooth devices for MindWave Mobile address"""
        if self.address is None:
            ut.log('info', "Scanning bluetooth devices...", self.verbose)
            try:
                devices = {v: k for k, v in
                           bluetooth.discover_devices(lookup_names=True)}
            except OSError:
                ut.log('error',
                       "Could not use bluetooth. Check bluetooth is on.",
                       self.verbose)
                sys.exit(1)
            else:
                if 'MindWave Mobile' not in devices:
                    ut.log('error',
                           "Could not find MindWave Mobile device. "
                           "Check that headset is on.", self.verbose)
                    sys.exit(1)
                else:
                    self.address = devices['MindWave Mobile']

    def connect(self):
        """Establishes connection with MindWave Mobile"""
        if self.socket is None:
            try:
                ut.log('info',
                       f"Connecting to MindWave Mobile "
                       f"({self.address})...",
                       self.verbose)
                self.socket = bluetooth.BluetoothSocket()
                self.socket.connect((self.address, 1))
            except bluetooth.btcommon.BluetoothError as e:
                ut.log('error',
                       f"Could not connect to MindWave Mobile: "
                       f"{e.strerror}", self.verbose)
                sys.exit(1)

    def start_reader(self):
        """Starts reader thread"""
        if self.thread is not None and self.thread.is_alive():
            ut.log('warn', "Background thread already started.", self.verbose)
        else:
            self.flag.clear()
            self.thread = MindWaveReader(self._data, self.callbacks,
                                         self.flag, self.socket,
                                         self.verbose)
            self.thread.start()

    def start(self):
        """Run 3 steps connection: scan, connect and start_reader
        This function simplifies the execution, not required"""
        self.scan()
        self.connect()
        self.start_reader()

    def stop(self):
        """Stops the reader thread"""
        if self.thread is not None and self.thread.is_alive():
            self.flag.set()
            self.thread.join()
            self.thread = None
            self.socket.close()
            self.socket = None

    def update_callback(self, target, callback):
        """Define callback function for a given target
        Executes a function when a value is updated

        :param target: The value to be tracked
        :type target: str
        :param callback: A function handle of the form ``callback(data)``,
                         where data is an `int` containing the updated value.
        :type callback: function"""
        self.callbacks[target] = callback

    def received(self):
        """Total packets received (and valid)"""
        return self.data['packets']

    def data(self, name):
        """Current value of variable
        :param name: Name of variable to request.
        :type name: str. Allowed values: signal, attention, meditation,
                         raw, eeg
        :return: Value of requested variable
        :rtype: int
        """
        if name in self._data['values']:
            return self._data['values'][name]
        else:
            ut.log('error', f"Name must be: {ut.NAMES}", self.verbose)
            return None





# SPDX-License-Identifier: GPL-3.0-or-later

# neuropy3 - neuropy3 backend.

# Copyright (C) 2022 Sergio Chica Manjarrez @ pervasive.it.uc3m.es.
# Universidad Carlos III de Madrid.

# This file is part of neuropy3.

# neuropy3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# neuropy3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with GNU Emacs.  If not, see <https://www.gnu.org/licenses/>.


from neuropy3.neuropy3 import MindWave
from neuropy3.gui import gui

import argparse


def main():
    files = {
        'raw': None,
        'att': None,
        'med': None,
        'eeg': None
    }

    def log_raw(data):
        # Approx. one packet every 2ms
        files['raw'].write(f'{data}\n')
        files['raw'].flush()

    def log_att(data):
        # Approx. one packet every 1s
        files['att'].write(f'{data}\n')
        files['att'].flush()

    def log_med(data):
        # Approx. one packet every 1s
        files['med'].write(f'{data}\n')
        files['med'].flush()

    def log_eeg(data):
        # Approx. one packet every 1s
        files['eeg'].write(f'{",".join(map(str, data.values()))}\n')
        files['eeg'].flush()

    parser = argparse.ArgumentParser(
        prog='python -m neuropy3',
        description=("NeuroSky MindWave Mobile 2 reader."))
    parser.add_argument('-v', '--verbose',
                        choices=range(5), default=2, type=int,
                        help="Maximum verbose level.")
    parser.add_argument('-r', '--raw', metavar='RAW_FILE',
                        nargs='?',  const='raw.csv',
                        help="Stores raw data in RAW_FILE. Default: raw.csv")
    parser.add_argument('-a', '--att', metavar='ATT_FILE',
                        nargs='?',  const='att.csv',
                        help=("Stores attention data in ATT_FILE. "
                              "Default: raw.csv)"))
    parser.add_argument('-m', '--med', metavar='MED_FILE',
                        nargs='?',  const='med.csv',
                        help=("Stores meditation data in MED_FILE. "
                              "Default: raw.csv)"))
    parser.add_argument('-e', '--eeg', metavar='EEG_FILE',
                        nargs='?',  const='eeg.csv',
                        help="Stores eeg data in EEG_FILE. Default: eeg.csv")
    parser.add_argument('-g', '--gui',
                        action='store_true',
                        help="Graphical interface to represent headset data.")
    parser.add_argument('-d', '--address',
                        metavar='bd_address',
                        help="MindWave Mobile bluetooth device address (MAC).")
    args = parser.parse_args()

    if not args.gui:
        mw = MindWave(args.address, autostart=False, verbose=args.verbose)
        if args.raw is not None:
            files['raw'] = open(args.raw, 'w')
            mw.update_callback('raw', log_raw)
        if args.att is not None:
            files['att'] = open(args.att, 'w')
            mw.update_callback('attention', log_att)
        if args.med is not None:
            files['med'] = open(args.med, 'w')
            mw.update_callback('meditation', log_med)
        if args.eeg is not None:
            files['eeg'] = open(args.eeg, 'w')
            mw.update_callback('eeg', log_eeg)
        mw.start()
        if mw.thread is not None:
            try:
                mw.thread.join()
            except KeyboardInterrupt:
                mw.stop()
                for file in files.values():
                    if file is not None:
                        file.close()
    else:
        gui.main(args.address)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-3.0-or-later

# setup - package setup.

# Copyright (C) 2022 Sergio Chica Manjarrez @ pervasive.it.uc3m.es.
# Universidad Carlos III de Madrid.

# This file is part of neuropy3.

# neuropy3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# neuropy3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with GNU Emacs.  If not, see <https://www.gnu.org/licenses/>.

from setuptools import setup
from pathlib import Path

_parent = Path(__file__).parent.resolve()
readme = (_parent / 'README.md').read_text(encoding='utf-8')


setup(
    name='neuropy3',
    version='1.0.0',
    description=('Python3 library to read data from '
                 'Neurosky Mindwave Mobile 2 in linux.'),
    author='Sergio Chica',
    author_email='sergio.chica@uc3m.es',
    long_description=readme,
    long_description_content_type='text/markdown',
    keywords=[
        'eeg',
        'neurosky',
        'mindwave',
        'neurosky mindwave mobile 2'
    ],
    url='https://gitlab.gast.it.uc3m.es/schica/neuropy3',
    license='GPLv3+',
    license_files=['LICENSE'],
    classifiers=[
        ('License :: OSI Approved :: '
         'GNU General Public License v3 or later (GPLv3+)'),
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8'
    ],
    packages=['neuropy3', 'neuropy3.gui'],
    python_requires='>= 3.8',
    install_requires=[
        'PySide6==6.2.3',
        'shiboken6==6.2.3',
        'pybluez2==0.46',
        'numpy==1.23.3',
        'scipy==1.9.1'
    ],
    entry_points={
        'console_scripts': [
            'neuropy=neuropy3.__main__:main'
        ]
    }
)



#!/usr/bin/env python3

# SPDX-License-Identifier: GPL-3.0-or-later

# neuropy3 - Utilities module.

# Copyright (C) 2022 Sergio Chica Manjarrez @ pervasive.it.uc3m.es.
# Universidad Carlos III de Madrid.

# This file is part of neuropy3.

# neuropy3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# neuropy3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with GNU Emacs.  If not, see <https://www.gnu.org/licenses/>.

from scipy.fft import irfft, rfft, rfftfreq

import numpy as np


_RED = '\033[91m'
_YELLOW = '\033[93m'
_BLUE = '\033[94m'
_GREEN = '\033[92m'
_CLEANC = '\033[0m'
_NC = ''

LOG = {
    'normal': '[*] ',
    'succ': '[+] ', 'info': '[*] ',
    'warn': '[-] ', 'error': '[!] '
}

BYTE = {
    'sync': b'\xaa',
    'step1': b'\xba',
    'step2': b'\xbc',
    'signal': b'\x02',
    'attention': b'\x04',
    'meditation': b'\x05',
    'raw': b'\x80',
    'eeg': b'\x83',
    '_max': b'\x7f'
}
CODE = {v: k for k, v in BYTE.items()}
NAMES = ['battery', 'signal', 'attention', 'meditation', 'blink',
         'raw', 'delta', 'theta', 'alpha_l', 'alpha_h',
         'beta_l', 'beta_h', 'gamma_l', 'gamma_m']
SAMPLE_RATE = 512
PLENGTH_MAX = 170
NO_CONTACT = 200
PKT_EEG_MAX = 24
PKT_RAW_MAX = 2
EEG = {
    'delta': (1, 4),
    'theta': (4, 8),
    'alpha': (8, 13),
    'beta': (13, 31),
    'gamma': (31, 51)
}  # Closed-open interval -> [)
WINDOW = None


def disable_ansi_colors():
    """Disable ANSI colors in log messages."""
    global _GREEN, _BLUE, _YELLOW, _RED, _CLEANC
    _GREEN = _BLUE = _YELLOW = _RED = _CLEANC = _NC


def set_logger(window):
    global WINDOW
    if WINDOW is None:
        WINDOW = window
        disable_ansi_colors()


def log(ltype, msg, level):
    """Print information, warning or error messages to stdout.
    :param ltype: Type of message. Allowed values: succ, info, warn or error.
    :type ltype: str
    :param msg: Message to print.
    :type msg: str
    :param level: Maximum level to print messages. Allowed values: 0-4
    :type level: int"""
    color = LOG[ltype]
    pmsg = False
    if ltype == 'succ':
        color = f'{_GREEN}{color}{_CLEANC}'
        if level > 3:
            pmsg = True
    elif ltype == 'warn':
        color = f'{_YELLOW}{color}{_CLEANC}'
        if level > 2:
            pmsg = True
    elif ltype == 'info':
        color = f'{_BLUE}{color}{_CLEANC}'
        if level > 1:
            pmsg = True
    elif ltype == 'error':
        color = f'{_RED}{color}{_CLEANC}'
        if level:
            pmsg = True

    if pmsg:
        if WINDOW is not None:
            WINDOW.emit(f"{color}{msg}")
        else:
            print(f"{color}{msg}")


def raw_to_microvolt(value):
    return round(((value * (1.8 / 4096)) / 2000) * 1e6, 3)


def microvolts_to_bands(microvolts):
    data_fft = rfft(microvolts)
    data_freq = rfftfreq(len(microvolts), d=1/SAMPLE_RATE)
    bands = []
    # Set irrelevant frequencies to 0, i.e. only frequencies in range
    # have values
    for band in EEG:
        idx = np.where(np.logical_or(data_freq < EEG[band][0],
                                     data_freq >= EEG[band][1]))
        tmp = data_fft.copy()
        tmp[idx] = 0
        inv = irfft(tmp)
        bands.append(list(inv))
    return bands


def signal_axes(signal):
    return min(signal), max(signal)
